package com.unciv.ui.objectdescriptions

import com.unciv.models.ruleset.unique.IHasUniques
import com.unciv.models.ruleset.unique.Unique
import com.unciv.models.ruleset.unique.UniqueType
import com.unciv.models.translations.tr
import com.unciv.ui.screens.civilopediascreen.FormattedLine

/**
 *  Appends user-visible Uniques as translated text to a [line collection][lineList].
 *
 *  Follows json order.
 *  @param exclude Predicate that can exclude Uniques by returning `true` (defaults to return `false`).
 */
fun IHasUniques.uniquesToDescription(
    lineList: MutableCollection<String>,
    exclude: Unique.() -> Boolean = {false}
): Unit {
    for (unique in uniqueObjects) {
        if (unique.isHiddenToUsers()) continue
        if (unique.exclude()) continue
        lineList += unique.text.tr()
    }
}

/**
 *  A Sequence of user-visible Uniques as [FormattedLine]s.
 *
 *  @param leadingSeparator Tristate: If there are lines to display and this parameter is not `null`, a leading line is output, as separator or empty line.
 *  @param sorted If set, sorts alphabetically (**not** using a locale-specific Collator). Otherwise lists in json order.
 *  @param colorConsumesResources If set, ConsumesResources Uniques get a reddish color.
 *  @param exclude Predicate that can exclude Uniques by returning `true` (defaults to return `false`).
 */
fun IHasUniques.uniquesToCivilopediaTextLines(
    leadingSeparator: Boolean? = false,
    sorted: Boolean = false,
    colorConsumesResources: Boolean = false,
    exclude: Unique.() -> Boolean = {false}
) = sequence {
    var orderedUniques = uniqueObjects.asSequence()
        .filterNot { it.isHiddenToUsers() || it.exclude() }
    if (sorted) orderedUniques = orderedUniques.sortedBy { it.text }

    for ((index, unique) in orderedUniques.withIndex()) {
        if (leadingSeparator != null && index == 0)
            yield(FormattedLine(separator = leadingSeparator))
        // Optionally special-case ConsumesResources to give it a reddish color. Also ensures link always points to the resource
        // (the other constructor guesses the first object by name in the Unique parameters).
        yield(
            if (colorConsumesResources && unique.isOfType(UniqueType.ConsumesResources))
                FormattedLine(unique.text, link = "Resources/${unique.params[1]}", color = "#F42")
                else FormattedLine(unique)
        )
    }
}

/**
 *  Appends user-visible Uniques as [FormattedLine]s to [lineList].
 *
 *  @param leadingSeparator Tristate: If there are lines to display and this parameter is not `null`, a leading line is output, as separator or empty line.
 *  @param sorted If set, sorts alphabetically (**not** using a locale-specific Collator). Otherwise lists in json order.
 *  @param colorConsumesResources If set, ConsumesResources Uniques get a reddish color.
 *  @param exclude Predicate that can exclude Uniques by returning `true` (defaults to return `false`).
 */
fun IHasUniques.uniquesToCivilopediaTextLines(
    lineList: MutableCollection<FormattedLine>,
    leadingSeparator: Boolean? = false,
    sorted: Boolean = false,
    colorConsumesResources: Boolean = false,
    exclude: Unique.() -> Boolean = {false}
) {
    uniquesToCivilopediaTextLines(leadingSeparator, sorted, colorConsumesResources, exclude)
        .toCollection(lineList)
}
